/* ---------------------------------------------------------------------
 *
 * -- Automatically Tuned Linear Algebra Software (ATLAS)
 *    (C) Copyright 1999 All Rights Reserved
 *
 * -- ATLAS routine -- Version 2.0 -- December 25, 1999
 *
 * -- Suggestions,  comments,  bugs reports should be sent to the follo-
 *    wing e-mail address: atlas@cs.utk.edu
 *
 *  Author         : Antoine P. Petitet
 * University of Tennessee - Innovative Computing Laboratory
 * Knoxville TN, 37996-1301, USA.
 *
 * ---------------------------------------------------------------------
 *
 * -- Copyright notice and Licensing terms:
 *
 * Redistribution  and  use in  source and binary forms, with or without
 * modification, are  permitted provided  that the following  conditions
 * are met:
 *
 * 1) Redistributions  of  source  code  must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2) Redistributions in binary form must reproduce  the above copyright
 *    notice,  this list of  conditions and the  following disclaimer in
 *    the documentation and/or other materials provided with the distri-
 *    bution.
 * 3) All advertising materials mentioning features or use of this soft-
 *    ware must display the folowing acknowledgement:
 *    This product includes software developed by the ATLAS group of the
 *    University of Tennesee, Knoxville and its contributors.
 * 4) The names of the  University of Tennessee,  Knoxville,  the  ATLAS
 *    group, or the names of its contributors may not be used to endorse
 *    or  promote products derived  from  this software without specific
 *    prior written permission.
 *
 * -- Disclaimer:
 *
 * The University of Tennessee, Knoxville,  the ATLAS group,  or the au-
 * thors make  no representations about the suitability of this software
 * for any purpose.  This software is provided ``as is'' without express
 * or implied warranty.
 *
 * ---------------------------------------------------------------------
 */
/*
 * Include files
 */
#include "atlas_refmisc.h"
#include "atlas_reflevel2.h"

void ATL_zrefher
(
   const enum ATLAS_UPLO      UPLO,
   const int                  N,
   const double               ALPHA,
   const double               * X,
   const int                  INCX,
   double                     * A,
   const int                  LDA
)
{
/*
 * Purpose
 * =======
 *
 * ATL_zrefher performs the Hermitian rank 1 operation
 *
 *    A := alpha* x * conjg( x' ) + A,
 *
 * where  alpha is a real scalar, x is an n-element vector and A is an n
 * by n Hermitian matrix.
 *
 * Arguments
 * =========
 *
 * UPLO    (input)                       const enum ATLAS_UPLO
 *         On entry, UPLO  specifies whether the upper or lower triangu-
 *         lar part of the array A is to be referenced as follows:
 *
 *             UPLO = AtlasUpper   Only the upper triangular part of A
 *                                 is to be referenced.
 *
 *             UPLO = AtlasLower   Only the lower triangular part of A
 *                                 is to be referenced.
 *
 *         Unchanged on exit.
 *
 * N       (input)                       const int
 *         On entry, N specifies the order of the matrix A. N must be at
 *         least zero. Unchanged on exit.
 *
 * ALPHA   (input)                       const double
 *         On entry, ALPHA specifies the real scalar alpha.  When  ALPHA
 *         is supplied as  zero then  the array X need not be set on in-
 *         put. Unchanged on exit.
 *
 * X       (input)                       const double *
 *         On entry,  X  points to the  first entry to be accessed of an
 *         incremented array of size equal to or greater than
 *            ( 1 + ( n - 1 ) * abs( INCX ) ) * sizeof( double[2] ),
 *         that contains the vector x. Unchanged on exit.
 *
 * INCX    (input)                       const int
 *         On entry, INCX specifies the increment for the elements of X.
 *         INCX must not be zero. Unchanged on exit.
 *
 * A       (input/output)                double *
 *         On entry,  A  points  to an array of size equal to or greater
 *         than   LDA * n * sizeof( double[2] ).   Before   entry   with
 *         UPLO = AtlasUpper, the leading n by n  upper  triangular part
 *         of the array A must contain the upper triangular  part of the
 *         Hermitian matrix  and the strictly lower triangular part of A
 *         is not referenced.  On exit, the upper triangular part of the
 *         array  A  is overwritten  by the upper triangular part of the
 *         updated matrix.  With UPLO = AtlasLower, the  leading  n by n
 *         lower triangular part of the array  A  must contain the lower
 *         triangular part of the Hermitian  matrix and the strictly up-
 *         per triangular part of A is not referenced.
 *         Note that  the  imaginary parts of the diagonal elements need
 *         not be set, they are assumed to be zero, and on exit they are
 *         set to zero.
 *
 * LDA     (input)                       const int
 *         On entry, LDA  specifies the leading dimension of A as decla-
 *         red  in  the  calling  (sub) program.  LDA  must be  at least
 *         MAX( 1, n ). Unchanged on exit.
 *
 * ---------------------------------------------------------------------
 */
/*
 * .. Local Variables ..
 */
   int                        i, iaij, incx2 = 2 * INCX, ix, j, jaj, jx,
                              lda2 = LDA << 1, ldap12 = ( LDA + 1 ) << 1;
   register double            t0_i, t0_r;
/* ..
 * .. Executable Statements ..
 *
 */
   if( ( N == 0 ) || ( ALPHA == ATL_dZERO ) ) return;

   if( UPLO == AtlasUpper )
   {
      for( j = 0,      jaj  = 0,    jx  = 0;
           j < N; j++, jaj += lda2, jx += incx2 )
      {
         Mset( ALPHA * X[jx], -ALPHA * X[jx+1], t0_r, t0_i );
         for( i = 0,      iaij  = jaj, ix  = 0;
              i < j; i++, iaij += 2,   ix += incx2 )
         {
            Mmla( X[ix], X[ix+1], t0_r, t0_i, A[iaij], A[iaij+1] );
         }
         Mset( A[iaij] + X[jx] * t0_r - X[jx+1] * t0_i, ATL_dZERO,
               A[iaij], A[iaij+1] );
      }
   }
   else
   {
      for( j = 0,      jaj  = 0,      jx  = 0;
           j < N; j++, jaj += ldap12, jx += incx2 )
      {
         Mset( ALPHA * X[jx], -ALPHA * X[jx+1], t0_r, t0_i );
         Mset( A[jaj] + X[jx] * t0_r - X[jx+1] * t0_i, ATL_dZERO,
               A[jaj], A[jaj+1] );
         for( i = j+1,    iaij  = jaj+2, ix  = jx+incx2;
              i < N; i++, iaij += 2,     ix += incx2 )
         {
            Mmla( X[ix], X[ix+1], t0_r, t0_i, A[iaij], A[iaij+1] );
         }
      }
   }
/*
 * End of ATL_zrefher
 */
}
