/* ---------------------------------------------------------------------
 *
 * -- Automatically Tuned Linear Algebra Software (ATLAS)
 *    (C) Copyright 1999 All Rights Reserved
 *
 * -- ATLAS routine -- Version 2.0 -- December 25, 1999
 *
 * -- Suggestions,  comments,  bugs reports should be sent to the follo-
 *    wing e-mail address: atlas@cs.utk.edu
 *
 *  Author         : Antoine P. Petitet
 * University of Tennessee - Innovative Computing Laboratory
 * Knoxville TN, 37996-1301, USA.
 *
 * ---------------------------------------------------------------------
 *
 * -- Copyright notice and Licensing terms:
 *
 * Redistribution  and  use in  source and binary forms, with or without
 * modification, are  permitted provided  that the following  conditions
 * are met:
 *
 * 1) Redistributions  of  source  code  must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2) Redistributions in binary form must reproduce  the above copyright
 *    notice,  this list of  conditions and the  following disclaimer in
 *    the documentation and/or other materials provided with the distri-
 *    bution.
 * 3) All advertising materials mentioning features or use of this soft-
 *    ware must display the folowing acknowledgement:
 *    This product includes software developed by the ATLAS group of the
 *    University of Tennesee, Knoxville and its contributors.
 * 4) The names of the  University of Tennessee,  Knoxville,  the  ATLAS
 *    group, or the names of its contributors may not be used to endorse
 *    or  promote products derived  from  this software without specific
 *    prior written permission.
 *
 * -- Disclaimer:
 *
 * The University of Tennessee, Knoxville,  the ATLAS group,  or the au-
 * thors make  no representations about the suitability of this software
 * for any purpose.  This software is provided ``as is'' without express
 * or implied warranty.
 *
 * ---------------------------------------------------------------------
 */
/*
 * Include files
 */
#include "atlas_refmisc.h"
#include "atlas_reflevel2.h"

void ATL_zreftpsv
(
   const enum ATLAS_UPLO      UPLO,
   const enum ATLAS_TRANS     TRANS,
   const enum ATLAS_DIAG      DIAG,
   const int                  N,
   const double               * A,
   double                     * X,
   const int                  INCX
)
{
/*
 * Purpose
 * =======
 *
 * ATL_zreftpsv solves one of the systems of equations
 *
 *    A * x = b,   or   conjg( A  ) * x = b,   or
 *
 *    A'* x = b,   or   conjg( A' ) * x = b,
 *
 * where b and x are n-element vectors and  A is an n by n unit, or non-
 * unit, upper or lower triangular matrix, supplied in packed form.
 *
 * No test for  singularity  or  near-singularity  is included  in  this
 * routine. Such tests must be performed before calling this routine.
 *
 * Arguments
 * =========
 *
 * UPLO    (input)                       const enum ATLAS_UPLO
 *         On entry, UPLO  specifies whether  the  matrix is an upper or
 *         lower triangular matrix as follows:
 *
 *             UPLO = AtlasUpper   A is an upper triangular matrix.
 *
 *             UPLO = AtlasLower   A is a lower triangular matrix.
 *
 *         Unchanged on exit.
 *
 * TRANS   (input)                       const enum ATLAS_TRANS
 *         On entry,  TRANS specifies the equations to be solved as fol-
 *         lows:
 *
 *            TRANS = AtlasNoTrans     A  * x = b,
 *
 *            TRANS = AtlasConj        conjg( A  ) * x = b,
 *
 *            TRANS = AtlasTrans       A' * x = b,
 *
 *            TRANS = AtlasConjTrans   conjg( A' ) * x = b.
 *
 *         Unchanged on exit.
 *
 * DIAG    (input)                       const enum ATLAS_DIAG
 *         On entry, DIAG specifies whether or not A is unit triangu-
 *         lar as follows:
 *
 *            DIAG = AtlasUnit       A is assumed to be unit triangular,
 *
 *            DIAG = AtlasNonUnit    A is not assumed to be unit trian-
 *                                   gular.
 *
 *         Unchanged on exit.
 *
 * N       (input)                       const int
 *         On entry, N specifies the order of the matrix A. N must be at
 *         least zero. Unchanged on exit.
 *
 * A       (input)                       const double *
 *         On entry,  A  points  to an array of size equal to or greater
 *         than   (( n*(n+1) ) / 2) * sizeof( double[2] ).  Before entry
 *         with UPLO = AtlasUpper, the array  A  must  contain the upper
 *         triangular matrix packed sequentially, column by  column,  so
 *         that A[ 0 ] contains a(0,0), A[ 1 ] and A[ 2 ] contain a(0,1)
 *         and  a(1,1)  respectively,  and  so  on.  Before  entry  with
 *         UPLO = AtlasLower, the array  A  must contain the  lower tri-
 *         angular matrix packed sequentially, column by column, so that
 *         A[ 0 ] contains a(0,0), A[ 1 ] and A[ 2 ] contain a(1,0)  and
 *         a( 2, 0 ) respectively, and so on.
 *
 *         Note that when  DIAG = AtlasUnit,  the diagonal elements of A
 *         are not referenced,  but are  assumed to be unity.  Unchanged
 *         on exit.
 *
 * X       (input/output)                double *
 *         On entry,  X  points to the  first entry to be accessed of an
 *         incremented array of size equal to or greater than
 *            ( 1 + ( n - 1 ) * abs( INCX ) ) * sizeof( double[2] ),
 *         that contains the vector x. Before entry, the incremented ar-
 *         ray X must contain the n element right-hand side vector b. On
 *         exit, X is overwritten with the solution vector x.
 *
 * INCX    (input)                       const int
 *         On entry, INCX specifies the increment for the elements of X.
 *         INCX must not be zero. Unchanged on exit.
 *
 * ---------------------------------------------------------------------
 */
/*
 * .. Local Variables ..
 */
   int                        i, iaij, iajj, incx2 = 2 * INCX, ix, j, jx;
   register double            t0_i, t0_r;
/* ..
 * .. Executable Statements ..
 *
 */
   if( N == 0 ) return;

   if( TRANS == AtlasNoTrans )
   {
      if( UPLO == AtlasUpper )
      {
         for( j = N-1,     iaij = (N-1)*N, jx  = (N-1)*incx2;
              j >= 0; j--,                 jx -= incx2 )
         {
            if( DIAG == AtlasNonUnit )
            {
               iajj = ( j << 1 ) + iaij;
               Mddiv( A[iajj], A[iajj+1], X[jx], X[jx+1] );
            }
            Mset( X[jx], X[jx+1], t0_r, t0_i );
            for( i = 0,                 ix  = 0;
                 i < j; i++, iaij += 2, ix += incx2 )
            {
               Mmls( A[iaij], A[iaij+1], t0_r, t0_i, X[ix], X[ix+1] );
            }
            iaij -= ( j << 2 );
         }
      }
      else
      {
         for( j = 0,      iaij = 0, jx  = 0;
              j < N; j++,           jx += incx2 )
         {
            if( DIAG == AtlasNonUnit )
               Mddiv( A[iaij], A[iaij+1], X[jx], X[jx+1] );
            iaij    += 2;
            Mset( X[jx], X[jx+1], t0_r, t0_i );
            for( i = j+1,               ix  = jx+incx2;
                 i < N; i++, iaij += 2, ix += incx2 )
            {
               Mmls( A[iaij], A[iaij+1], t0_r, t0_i, X[ix], X[ix+1] );
            }
         }
      }
   }
   else if( TRANS == AtlasConj )
   {
      if( UPLO == AtlasUpper )
      {
         for( j = N-1,     iaij = (N-1)*N, jx  = (N-1)*incx2;
              j >= 0; j--,                 jx -= incx2 )
         {
            if( DIAG == AtlasNonUnit )
            {
               iajj = ( j << 1 ) + iaij;
               Mddiv( A[iajj], -A[iajj+1], X[jx], X[jx+1] );
            }
            Mset( X[jx], X[jx+1], t0_r, t0_i );
            for( i = 0,                 ix  = 0;
                 i < j; i++, iaij += 2, ix += incx2 )
            {
               Mmls( A[iaij], -A[iaij+1], t0_r, t0_i, X[ix], X[ix+1] );
            }
            iaij -= ( j << 2 );
         }
      }
      else
      {
         for( j = 0,      iaij = 0, jx  = 0;
              j < N; j++,           jx += incx2 )
         {
            if( DIAG == AtlasNonUnit )
               Mddiv( A[iaij], -A[iaij+1], X[jx], X[jx+1] );
            iaij += 2;
            Mset( X[jx], X[jx+1], t0_r, t0_i );
            for( i = j+1,               ix  = jx+incx2;
                 i < N; i++, iaij += 2, ix += incx2 )
            {
               Mmls( A[iaij], -A[iaij+1], t0_r, t0_i, X[ix], X[ix+1] );
            }
         }
      }
   }
   else if( TRANS == AtlasTrans )
   {
      if( UPLO == AtlasUpper )
      {
         for( j = 0,      iaij = 0, jx  = 0;
              j < N; j++,           jx += incx2 )
         {
            Mset( X[jx], X[jx+1], t0_r, t0_i );
            for( i = 0,                 ix  = 0;
                 i < j; i++, iaij += 2, ix += incx2 )
            {
               Mmls( A[iaij], A[iaij+1], X[ix], X[ix+1], t0_r, t0_i );
            }
            if( DIAG == AtlasNonUnit ) Mddiv( A[iaij], A[iaij+1], t0_r, t0_i );
            Mset( t0_r, t0_i, X[jx], X[jx+1] );
            iaij   += 2;
         }
      }
      else
      {
         for( j = N-1,     iajj = ((N-1)*(N+2)), jx  = (N-1)*incx2;
              j >= 0; j--,                       jx -= incx2 )
         {
            Mset( X[jx], X[jx+1], t0_r, t0_i );
            for( i = j+1,    iaij  = iajj+2, ix  = jx + incx2;
                 i < N; i++, iaij += 2,      ix += incx2 )
            {
               Mmls( A[iaij], A[iaij+1], X[ix], X[ix+1], t0_r, t0_i );
            }
            if( DIAG == AtlasNonUnit ) Mddiv( A[iajj], A[iajj+1], t0_r, t0_i );
            Mset( t0_r, t0_i, X[jx], X[jx+1] );
            iajj   -= ( N - j + 1 ) << 1;
         }
      }
   }
   else if( TRANS == AtlasConjTrans )
   {
      if( UPLO == AtlasUpper )
      {
         for( j = 0,      iaij = 0, jx  = 0;
              j < N; j++,           jx += incx2 )
         {
            Mset( X[jx], X[jx+1], t0_r, t0_i );
            for( i = 0,                 ix  = 0;
                 i < j; i++, iaij += 2, ix += incx2 )
            {
               Mmls( A[iaij], -A[iaij+1], X[ix], X[ix+1], t0_r, t0_i );
            }
            if( DIAG == AtlasNonUnit )
               Mddiv( A[iaij], -A[iaij+1], t0_r, t0_i );
            Mset( t0_r, t0_i, X[jx], X[jx+1] );
            iaij += 2;
         }
      }
      else
      {
         for( j = N-1,     iajj = ((N-1)*(N+2)), jx  = (N-1)*incx2;
              j >= 0; j--,                       jx -= incx2 )
         {
            Mset( X[jx], X[jx+1], t0_r, t0_i );
            for( i = j+1,    iaij  = iajj+2, ix  = jx + incx2;
                 i < N; i++, iaij += 2,      ix += incx2 )
            {
               Mmls( A[iaij], -A[iaij+1], X[ix], X[ix+1], t0_r, t0_i );
            }
            if( DIAG == AtlasNonUnit )
               Mddiv( A[iajj], -A[iajj+1], t0_r, t0_i );
            Mset( t0_r, t0_i, X[jx], X[jx+1] );
            iajj -= ( N - j + 1 ) << 1;
         }
      }
   }
/*
 * End of ATL_zreftpsv
 */
}
